from jinja2 import nodes
from jinja2.ext import Extension
from .component import get_component_class

from jinja2_simple_tags import StandaloneTag


class MeldTag(StandaloneTag):
    tags = {"meld"}

    def render(self, component_name, **kwargs):
        mn = MeldNode(component_name)
        return mn.render(**kwargs)


class MeldScriptsTag(Extension):
    """
    Create a {% meld_scripts %} tag.
    Used to add the necessary js files to init meld
    """

    tags = {"meld_scripts"}

    def parse(self, parser):
        lineno = parser.stream.expect("name:meld_scripts").lineno

        call = self.call_method("_render", lineno=lineno)
        return nodes.Output([nodes.MarkSafe(call)]).set_lineno(lineno)

    def _render(self):
        files = ["morphdom-umd.js", "socket.io.js"]
        msg_url = "message"
        base_js_url = "/meld_js_src"
        scripts = ""
        for f in files:
            url = f"{base_js_url}/{f}"
            scripts += f'<script src="{url}"></script>'

        meld_url = f"{base_js_url}/meld.js"
        meld_import = f'import {{Meld}} from "{meld_url}";'
        scripts += f'<script type="module" src="{meld_url}"></script>'
        scripts += (
            '<script type="module">'
            f'var url = "{msg_url}"; {meld_import} Meld.init(url); </script>'
        )

        return scripts


class MeldNode:
    def __init__(self, component):
        self.component_name = component

    def render(self, **kwargs):
        Component = get_component_class(self.component_name)
        component = Component(**kwargs)
        rendered_component = component.render(self.component_name)

        return rendered_component



import ast
from werkzeug.wrappers.response import Response
import functools

from .component import get_component_class
from flask import jsonify, current_app
import orjson


def process_message(message):
    meld_id = message["id"]
    component_name = message["componentName"]
    action_queue = message["actionQueue"]

    data = message["data"]
    Component = get_component_class(component_name)
    component = Component(meld_id, **data)
    return_data = None

    for action in action_queue:
        payload = action.get("payload", None)
        if "syncInput" in action["type"]:
            if hasattr(component, payload["name"]):
                setattr(component, payload["name"], payload["value"])
                if component._form:
                    field_name = payload.get("name")
                    if field_name in component._form._fields:
                        field = getattr(component._form, field_name)
                        component._set_field_data(field_name, payload["value"])
                        component.updated(field)
                        component.errors[field_name] = field.errors or ""
                else:
                    component.updated(payload["name"])

        elif "callMethod" in action["type"]:
            call_method_name = payload.get("name", "")
            method_name, params = parse_call_method_name(call_method_name)
            message = payload.get("message")

            if method_name is not None and hasattr(component, method_name):
                func = getattr(component, method_name)
                if params:
                    return_data = func(*params)
                elif message:
                    return_data = func(**message)
                else:
                    return_data = func()
                if component._form:
                    component._bind_form(component._attributes())

    rendered_component = component.render(component_name)

    res = {
        "id": meld_id,
        "dom": rendered_component,
        "data": orjson.dumps(jsonify(component._attributes()).json).decode("utf-8"),
    }

    if type(return_data) is Response and return_data.status_code == 302:
        res["redirect"] = {"url": return_data.location}
    return res


def process_init(component_name):
    Component = get_component_class(component_name)
    return Component._listeners()


def parse_call_method_name(call_method_name: str):
    params = None
    method_name = call_method_name

    if "(" in call_method_name and call_method_name.endswith(")"):
        param_idx = call_method_name.index("(")
        params_str = call_method_name[param_idx:]

        # Remove the arguments from the method name
        method_name = call_method_name.replace(params_str, "")

        # Remove parenthesis
        params_str = params_str[1:-1]
        if params_str != "":
            try:
                params = ast.literal_eval("[" + params_str + "]")
            except ValueError:
                params = list(map(str.strip, params_str.split(",")))

    return method_name, params


def listen(*event_names: str):
    """
    Decorator to indicate that the decorated method should listen for custom events.
    It can be called using `flask_meld.emit`. Keyword arguments from `flask_meld.emit`
    will be passed as keyword arguments to the decorated method.

    Params:
        *event_names (str): One or more event names to listen for.
    """
    def dec(func):
        func._meld_event_names = event_names
        return func
    return dec


def emit(event_name: str, **kwargs):
    """
    Emit a custom event which will call any Component methods with the `@listen`
    decorator that are listening for the given event. Keyword arguments to this
    function are passed as keyword arguments to each of the decorated methods.

    Params:
        event_name (str): The name of the custom event to emit.
        **kwargs: Arguments to be passed as keyword arguments to the listening
            methods.
    """
    current_app.socketio.emit("meld-event", {"event": event_name, "message": kwargs})



import os
import uuid
from importlib.util import module_from_spec, spec_from_file_location
from itertools import groupby
from operator import itemgetter

import orjson
from bs4 import BeautifulSoup
from bs4.element import Tag
from bs4.formatter import HTMLFormatter
from flask import current_app, jsonify, render_template
from jinja2.exceptions import TemplateNotFound


def convert_to_snake_case(s):
    s.replace("-", "_")
    return s


def convert_to_camel_case(s):
    s = convert_to_snake_case(s)
    return "".join(word.title() for word in s.split("_"))


def get_component_class(component_name):
    """
    Get a component class based on a component name.
    """
    module_name = convert_to_snake_case(component_name)
    class_name = convert_to_camel_case(module_name)
    module = get_component_module(module_name)
    component_class = getattr(module, class_name)

    return component_class


def get_component_module(module_name):
    """
    Get the module from the meld/components directory or from a
    custom location using the config with `MELD_COMPONENT_DIR`.
    """
    user_specified_dir = current_app.config.get("MELD_COMPONENT_DIR", None)

    if not user_specified_dir:
        try:
            name = getattr(current_app, "name", None)
            full_path = os.path.join(name, "meld", "components", module_name + ".py")
            module = load_module_from_path(full_path, module_name)
        except FileNotFoundError:
            full_path = os.path.join("meld", "components", module_name + ".py")
            module = load_module_from_path(full_path, module_name)
        return module
    else:
        try:
            full_path = os.path.join(user_specified_dir, module_name + ".py")
            module = load_module_from_path(full_path, module_name)
        except FileNotFoundError:
            full_path = os.path.join(
                user_specified_dir, "components", module_name + ".py"
            )
            module = load_module_from_path(full_path, module_name)
        return module


def load_module_from_path(full_path, module_name):
    """
    Load a module given from a given path based on the name of the module
    """
    spec = spec_from_file_location(module_name, full_path)
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


CSRF_TOKEN_ATTR = "csrf_token"


class Component:
    """
    The meld Component class does most of the heavy lifting to handle data-binding,
    template context variable binding, template rendering and additional hooks.
    """

    def __init__(self, id=None, **kwargs):
        if not id:
            id = uuid.uuid4()
        self.errors = {}
        self._form = None
        self.__dict__.update(**kwargs)
        self.id = id

        if hasattr(self, "form"):
            self._bind_form(kwargs)

    def __repr__(self):
        return f"<meld.Component {self.__class__.__name__}>"

    @classmethod
    def _listeners(cls):
        """
        Dictionary containing all listeners and the methods they call
        """
        listeners = [
            (event_name, func.__name__)
            for func in cls.__dict__.values()
            if hasattr(func, "_meld_event_names")
            for event_name in func._meld_event_names
        ]
        return {
            event_name: [t[1] for t in group]
            for event_name, group in groupby(listeners, itemgetter(0))
        }

    @property
    def _meld_attrs(self):
        """
        A list of meld variables and functions that are hidden from the view
        """
        return ["id", "render", "validate", "updated", "form"]

    def _bind_form(self, kwargs):
        """
        Create a form from the form_class, add meld:model to each field and
        bind kwargs to field data.
        """
        self._form = getattr(self, "form")
        for field in self._form:
            if not field.type == "SubmitField":
                meld_attribute = {"meld:model": field.name}
                setattr(self._form[field.name], "render_kw", meld_attribute)
                self._bind_data_to_form(field, kwargs)

    def _set_token(self, field):
        """
        Gather the CSRF token from the form and apply it to the component.
        """
        soup = BeautifulSoup(field.__call__(), features="html.parser")
        token = soup.find(attrs={"name": CSRF_TOKEN_ATTR}).get("value")
        self.csrf_token = token

    def _bind_data_to_form(self, field, kwargs):
        """
        Bind any attributes from kwargs that are form fields to the form.
        """
        if field.name in kwargs:
            self._set_field_data(field.name, kwargs[field.name])
            if field.name == CSRF_TOKEN_ATTR:
                self._set_token(field)
        else:
            setattr(self, field.name, None)

    def _set_field_data(self, field_name, value):
        """
        Set the data attribute on a form field.
        """
        setattr(self._form[field_name], "data", value)

    def validate(self, field=None):
        """
        Validate a form or a field.
        """
        if not self._form:
            return True

        if field:
            validate = field.validate(self._form)
        else:
            validate = self._form.validate()

        if not validate:
            for field in self._form:
                if field.errors:
                    self.errors[field.name] = field.errors
        return validate

    def _attributes(self):
        """
        Get attributes that can be called in the component.
        """
        attributes = {}

        attributes_names = [
            attr
            for attr in dir(self)
            if not callable(getattr(self, attr))
            and not attr.startswith("_")
            and attr not in self._meld_attrs
        ]
        for name in attributes_names:
            attributes[name] = getattr(self, name)

        return attributes

    def _functions(self):
        """
        Get methods that can be called in the component.
        """

        functions = {}

        function_list = [
            func
            for func in dir(self)
            if callable(getattr(self, func))
            and not func.startswith("_")
            and func not in self._meld_attrs
        ]

        for func in function_list:
            functions[func] = getattr(self, func)

        return functions

    def __context__(self):
        """
        Collects every thing that could be used in the template context.
        """
        return {
            "attributes": self._attributes(),
            "methods": self._functions(),
        }

    def updated(self, name):
        """
        Hook that gets called when a component's data is about to get updated.
        """
        pass

    def render(self, component_name: str):
        return self._view(component_name)

    def _render_template(self, template_name: str, context_variables: dict):
        try:
            return render_template(template_name, **context_variables)
        except TemplateNotFound:
            return render_template(f"meld/{template_name}", **context_variables)

    def _view(self, component_name: str):
        data = self._attributes()
        context = self.__context__()
        context_variables = {}
        context_variables.update(context["attributes"])
        context_variables.update(context["methods"])
        context_variables.update({"form": self._form})

        rendered_template = self._render_template(
            f"{component_name}.html", context_variables
        )

        soup = BeautifulSoup(rendered_template, features="html.parser")
        root_element = Component._get_root_element(soup)
        root_element["meld:id"] = str(self.id)
        self._set_values(root_element, context_variables)

        script = soup.new_tag("script", type="module")
        init = {"id": str(self.id), "name": component_name, "data": jsonify(data).json}
        init_json = orjson.dumps(init).decode("utf-8")

        meld_import = 'import {Meld} from "/meld_js_src/meld.js";'
        script.string = f"{meld_import} Meld.componentInit({init_json});"
        root_element.append(script)

        rendered_template = Component._desoupify(soup)

        return rendered_template

    def _set_values(self, soup, context_variables):
        """
        Set the value on model fields
        """
        for element in soup.select("input,select,textarea"):
            model_attrs = [
                attr for attr in element.attrs.keys() if attr.startswith("meld:model")
            ]
            if len(model_attrs) > 1:
                raise Exception(
                    "Multiple 'meld:model' attributes not allowed on one tag."
                )

            for model_attr in model_attrs:
                value = context_variables[element.attrs[model_attr]]
                element.attrs["value"] = value
                if element.name == "select":
                    for e in element.find_all("option"):
                        if type(e) is Tag and e.attrs.get("value") == value:
                            e["selected"] = ""

                elif (
                    element.attrs.get("type")
                    and element.attrs.get("type") == "checkbox"
                    and value is True
                ):
                    element["checked"] = ""

    @staticmethod
    def _get_root_element(soup):
        for element in soup.contents:
            if element.name:
                return element

        raise Exception("No root element found")

    @staticmethod
    def _desoupify(soup):
        soup.smooth()
        return soup.encode(formatter=UnsortedAttributes()).decode("utf-8")


class UnsortedAttributes(HTMLFormatter):
    """
    Prevent beautifulsoup from re-ordering attributes.
    """

    def attributes(self, tag):
        for k, v in tag.attrs.items():
            yield k, v



from string import Template

requirements_template = Template(
    """
Flask>=0.9
Flask-Meld>=0.7.0
python-dotenv>=0.17.0
"""
)

config_template = Template(
    """
import os
import secrets


class Config:
    SECRET_KEY = os.environ.get('SECRET_KEY') or secrets.token_hex(16)


class DevelopmentConfig(Config):
    DEBUG = True


class ProductionConfig(Config):
    DEBUG = False


class TestingConfig(Config):
    TESTING = True


config = {
    'development': DevelopmentConfig,
    'testing': TestingConfig,
    'production': ProductionConfig,
}
"""
)

init_template = Template(
    """
from flask import Flask, render_template
from config import config
from flask_meld import Meld
# from .db import db
# from app import models

# extensions
meld = Meld()


def create_app(config_name="development"):
    app = Flask(__name__)
    app.config.from_object(config[config_name])
    # db.init_app(app)

    meld.init_app(app)

    @app.route("/")
    def index():
        return render_template("index.html")

    return app
    """
)

env_template = Template(
    """
SECRET_KEY=$secret_key
FLASK_ENV=development
"""
)

wsgi_template = Template(
    """
from app import create_app

app = create_app(config_name='production')
socketio = app.socketio


if __name__ == "__main__":
    socketio.run(app=app)
"""
)

base_html_template = Template(
    """
<html>
    <head>
        <title>Welcome to Flask-Meld</title>
        <meta charset="utf-8">
        <meta http-equiv="X-UA-Compatible" content="IE=edge">

        {% block head_scripts %}
        {% endblock %}
    </head>
    <body>
        {% meld_scripts %}

        {% block content %}

        {% endblock %}

        {% block page_scripts %}
        {% endblock %}
    </body>
</html>
"""
)

index_html_template = Template(
    """
{% extends "base.html" %}

{% block content %}
<h1>Flask-Meld</h1>
{% endblock %}
"""
)

components = Template(
    """
from flask_meld.component import Component


class $class_name(Component):
    example_variable = False

    def example_function(self):
        self.example_variable = not self.example_variable
"""
)

components_template = Template(
    """
<div>
    <button meld:click="example_function">Toggle</button>
    <input type="text" meld:model="example_variable" readonly></input>
</div>
"""
)



import os
import secrets
from pathlib import Path
from flask.cli import with_appcontext
from flask import current_app

import click

from flask_meld.templates import (
    base_html_template,
    config_template,
    components,
    components_template,
    env_template,
    index_html_template,
    init_template,
    requirements_template,
    wsgi_template,
)


@click.group()
def meld():
    """Flask-Meld specific commands"""


@meld.group()
def new():
    """Commands for new keyword"""


@new.command("project")
@click.argument("name")
def project(name):
    """Create a new flask-meld app with application defaults"""
    click.echo(f"Creating app {name}")
    generate_meld_app(name)


@new.command("component")
@click.argument("name")
@with_appcontext
def component(name):
    """Create a new component"""
    click.echo(f"Creating component '{name}'.")
    generate_meld_component(name)


def generate_meld_app(name):
    try:
        base_dir = Path.cwd() / name
        os.makedirs(base_dir / "app" / "meld" / "components")
        os.makedirs(base_dir / "app" / "meld" / "templates")
        os.makedirs(base_dir / "app" / "templates")
        os.makedirs(base_dir / "app" / "static" / "images")
        os.makedirs(base_dir / "app" / "static" / "css")
        os.makedirs(base_dir / "tests")
        generate_file_with_content(
            base_dir, "requirements.txt", requirements_template.template
        )
        generate_file_with_content(base_dir, "config.py", config_template.template)
        generate_file_with_content(base_dir, "app/__init__.py", init_template.template)
        generate_file_with_content(base_dir, "app/wsgi.py", wsgi_template.template)
        generate_file_with_content(
            base_dir, "app/templates/base.html", base_html_template.template
        )
        generate_file_with_content(
            base_dir, "app/templates/index.html", index_html_template.template
        )

        generated_secret_key = secrets.token_hex(16)
        generate_file_with_content(
            base_dir, ".env", env_template.substitute(secret_key=generated_secret_key)
        )
    except OSError:
        pass


def generate_meld_component(name):
    name = name.lower()
    try:
        base_dir = Path(current_app.root_path)
        components_dir = base_dir / "meld" / "components"
        templates_dir = base_dir / "meld" / "templates"

        if not (os.path.exists(components_dir) and os.path.exists(templates_dir)):
            click.echo(f"Failed. Could not find: {components_dir} or {templates_dir}")
            return False

        name_split = name.split("_")

        class_name = ""
        for name_seq in name_split:
            class_name += name_seq.capitalize()

        component = components_dir / f"{name}.html"
        if os.path.exists(component):
            click.echo(f"Failed. Component '{name}' already exists.")
            return False

        template = templates_dir / f"{name}.html"
        if os.path.exists(template):
            click.echo(f"Failed. Template '{template}' already exists.")
            return False

        generate_file_with_content(
            components_dir, f"{name}.py", components.substitute(class_name=class_name)
        )
        generate_file_with_content(
            templates_dir, f"{name}.html", components_template.template
        )
        click.echo(f"Component '{name}' successfully created.")

    except OSError:
        click.echo(
            "Failed. Unable to write to disk. Verify you have sufficient permissions."
        )
        return False


def generate_file_with_content(path, filename, file_contents):
    path = Path(f"{path}/{filename}")
    with open(path, "w") as f:
        f.write(file_contents)
    return path


if __name__ == "__main__":
    meld()



import os
from pathlib import Path

import jinja2
import pkg_resources
from flask import send_from_directory
from flask_socketio import SocketIO

from .message import process_init, process_message
from .tag import MeldScriptsTag, MeldTag


class Meld:
    def __init__(self, app=None, socketio=None, **kwargs):
        self.app = app

        if app is not None:
            self.init_app(app, socketio=socketio, **kwargs)

    def send_static_file(self, filename):
        """Send a static file from the flask-meld js directory."""
        directory = Path(pkg_resources.resource_filename('flask_meld', 'meld_js_src'))
        return send_from_directory(directory, filename)

    def init_app(self, app, socketio=None, **kwargs):
        app.jinja_env.add_extension(MeldTag)
        app.jinja_env.add_extension(MeldScriptsTag)

        # Load templates from template dir or app/meld/templates
        custom_template_loader = jinja2.ChoiceLoader([
            app.jinja_loader,
            jinja2.FileSystemLoader('app/meld/templates'),
        ])

        app.jinja_loader = custom_template_loader
        if socketio:
            app.socketio = socketio
        else:
            app.socketio = SocketIO(app, **kwargs)

        meld_dir = app.config.get("MELD_COMPONENT_DIR", None)
        if meld_dir:
            if not os.path.isabs(meld_dir):
                directory = os.path.abspath(os.path.join(app.root_path, meld_dir))
                app.config["MELD_COMPONENT_DIR"] = directory

        if not app.config.get("SECRET_KEY"):
            raise RuntimeError(
                "The Flask-Meld requires the 'SECRET_KEY' config " "variable to be set"
            )

        @app.route("/meld_js_src/<path:filename>")
        def meld_static_file(filename):
            return self.send_static_file(filename)

        @app.socketio.on("meld-message")
        def meld_message(message):
            """meldID, action, componentName"""
            result = process_message(message)
            app.socketio.emit("meld-response", result)

        @app.socketio.on("meld-init")
        def meld_init(message):
            return process_init(message)



"""
Flask-Meld
--------------
A way to meld your frontend and backend code
"""
import pathlib
from setuptools import setup

with open("flask_meld/__init__.py", "r") as f:
    version = [
        line.split(" = ")[1].strip()
        for line in f.readlines()
        if line.startswith("__version__")
    ][0].strip('"')

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(
    name="Flask-Meld",
    version=version,
    url="http://github.com/mikeabrahamsen/Flask-Meld/",
    license="MIT",
    author="Michael Abrahamsen",
    author_email="mail@michaelabrahamsen.com",
    description="Meld is a framework for Flask that allows you to create dynamic user interfaces using Python and the Jinja2 templating engine.",
    long_description=README,
    long_description_content_type="text/markdown",
    packages=["flask_meld"],
    zip_safe=False,
    include_package_data=True,
    platforms="any",
    install_requires=[
        "Flask>=0.9",
        "beautifulsoup4>=4",
        "orjson>=3.4.6",
        "flask-socketio>=5",
        "gevent-websocket>=0.10.1",
        "jinja2-simple-tags==0.3.1",
        "click==7.1.2"
    ],
    tests_require=["pytest"],
    test_suite="tests",
    classifiers=[
        "Environment :: Web Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)
